#include "ext.h"
#include "z_dsp.h"
#include "math.h"


#define DELAY_ALIGN	16		/* alignment needed, in bytes, for delay line */

typedef struct delbuf
{
	float *b_shit;
	float b_conv;
	long b_phase;
	long b_nsamps;
} t_delbuf;

typedef struct _delwrite
{
	t_pxobject x_obj;
	t_symbol *x_sym;
	struct _delwrite *x_next;	/* list of delwrites hanging from symbol */
	t_delbuf x_buf;
	char *x_malloc;
	long x_on;				// hack, has it been put on chain yet?
} t_delwrite;

void *delwrite_class;
t_symbol *ps_delwrite_class;

void cu_delwrite(float *in, t_delbuf *b, long n);
t_int *delwrite_perform(t_int *w);
void delwrite_dsp(t_delwrite *x, t_signal **sp, short *count);
void delwrite_assist(t_delwrite *x, void *b, long m, long a, char *s);
void delwrite_free(t_delwrite *x);
void *delwrite_new(t_symbol *s, double maxdel, long phys);

void main(void)
{
	setup((t_messlist **)&delwrite_class, (method)delwrite_new, (method)delwrite_free, (short)sizeof(t_delwrite), 
		0L, A_SYM, A_DEFFLOAT, A_DEFLONG, 0);
	addmess((method)delwrite_dsp, "dsp", A_CANT, 0);
	addmess((method)delwrite_assist, "assist", A_CANT, 0);
	dsp_initclass();
	ps_delwrite_class = gensym("delwrite_class");
	ps_delwrite_class->s_thing = (t_object *)&delwrite_class;
}

void cu_delwrite(float *in, t_delbuf *b, long n)
{
    float *bp = b->b_shit + b->b_phase;
    int i;
    for (i = 0; i < n; i += 4,  bp+= 4,  in += 4)
    {
      bp[0] = in[0] + 1E-20;
      bp[1] = in[1] + 1E-20;
      bp[2] = in[2] + 1E-20;
      bp[3] = in[3] + 1E-20;
    }
    /*
    for (i = 0; i < n; i ++)
    {
      *bp++ = *in++ + 1E-20;
    } */
    if (b->b_phase == b->b_nsamps)
    {
		sysmem_copyptr(b->b_shit + b->b_phase, b->b_shit, n * sizeof(float));
		b->b_phase = n;
    }
    else b->b_phase += n;
}

t_int *delwrite_perform(t_int *w)
{
	float *in;
	t_delwrite *x = (t_delwrite *)(w[1]);
	int n;
	
	x->x_on = 0;
	if (x->x_obj.z_disabled)
		goto out;
	in = (float *)(w[2]);
	n = w[3];
	
	cu_delwrite(in, &x->x_buf, n);
out:
	return w + 4;
}

void delwrite_dsp(t_delwrite *x, t_signal **sp, short *count)
{
	if (!x->x_malloc) {
		error("delwrite~ %s: no delay memory",x->x_sym->s_name);
		return;
	}
	if (!x->x_sym->s_thing)
		post(" warning: delwrite~ %s: something else had the same name", x->x_sym->s_name);
	else if (x->x_sym->s_thing != (t_object *)x)
		post(" warning: delwrite~ %s: multiply defined", x->x_sym->s_name);
	x->x_buf.b_conv = sp[0]->s_sr * .001;
	dsp_add(delwrite_perform, 3, x, sp[0]->s_vec, sp[0]->s_n);
	x->x_on = 1;
}

void delwrite_assist(t_delwrite *x, void *b, long m, long a, char *s)
{
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"(signal) Input Written To Delay Line"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"(dummy)"); break;
		}
	}
}

void delwrite_free(t_delwrite *x)
{
	dsp_free((t_pxobject *)x);
	if (x->x_sym->s_thing && ob_messlist(x->x_sym->s_thing) == delwrite_class) {
		t_delwrite stub, *y, *z;
		stub.x_next = (t_delwrite *)(x->x_sym->s_thing);
		for (y = &stub; z = y->x_next; y = z) {
			if (z == x) {
				y->x_next = z->x_next;
				break;
			}
		}
		x->x_sym->s_thing = (t_object *)(stub.x_next);
	}
	if (x->x_malloc)
		sysmem_freeptr((Ptr)x->x_malloc);	
}

/* problem -- we don't know the sampling rate yet!	How do you know
	 how big to make the buffer??		*/

void *delwrite_new(t_symbol *s, double maxdel, long phys)
{
	t_delwrite *x = (t_delwrite *)newobject(delwrite_class);
	long samps, length, i;
	float *fp;
	
	if (maxdel < 0)
		maxdel = 0;
	samps = SYS_MAXBLKSIZE + .001 * sys_getsr() * maxdel;
	if (samps < SYS_MAXBLKSIZE)
		samps = SYS_MAXBLKSIZE;
	samps += (-samps) & (SYS_MAXBLKSIZE - 1);
	length = samps + SYS_MAXBLKSIZE;

	/* just ignore the phys argument, and allocate the memory */
	x->x_malloc = (char *)sysmem_newptr(length * sizeof(float) + DELAY_ALIGN);
	if (x->x_malloc) {
		/* fix the alignment */
		x->x_buf.b_shit = (float *)((long)(x->x_malloc) +
					((-(long)(x->x_malloc)) & (DELAY_ALIGN - 1)));
	} else {
		error("delwrite~: no space for delay line");
		x->x_buf.b_shit = 0;
	}
	if (x->x_buf.b_shit) {
		for (i = length, fp = x->x_buf.b_shit; i--;)
			*(fp++) = 0;
	}
	
	x->x_buf.b_nsamps = samps;
	x->x_buf.b_phase = 0;
	dsp_setup((t_pxobject *)x, 1);
	x->x_sym = s;
	if (s->s_thing) {
		if (ob_messlist(s->s_thing) != delwrite_class) {
			error("%s: already exists and is not a delay line", s->s_name);
		} else {
			x->x_next = ((t_delwrite *)(s->s_thing));
			s->s_thing = (t_object *)x;
		}
	} else {
		s->s_thing = (t_object *)x;
		x->x_next = 0;
	}
	outlet_new((t_object *)x,"signal");
	return (x);
}

