#include "ext.h"
#include "z_dsp.h"
#include "math.h"


typedef struct twopole
{
	float c_ym1;		/* last output value of filt */
	float c_ym2;		/* second - to - last */
	float c_gain;		/* input gain */
	float c_coef1;		/* first feedback coefficient */
	float c_coef2;		/* second feedback coefficient */
} t_twopole;

typedef struct _sigwahwah
{
	t_pxobject x_obj;
	t_twopole x_2p;
	float x_fconv;
} t_sigwahwah;

void cu_do2pole(float *in, float *out, long n, t_twopole *x);
t_int *sigwahwah_perform(t_int *w);
void sigwahwah_dsp(t_sigwahwah *x, t_signal **sp, short *count);
void sigwahwah_assist(t_sigwahwah *x, void *b, long m, long a, char *s);
void *sigwahwah_new(long old);

void *sigwahwah_class;

void main(void)
{
	setup((t_messlist **)&sigwahwah_class, (method)sigwahwah_new, (method)dsp_free, (short)sizeof(t_sigwahwah), 
		0L, A_DEFLONG, 0);
	dsp_initclass();
	addmess((method)sigwahwah_dsp, "dsp", A_CANT, 0);
	addmess((method)sigwahwah_assist, "assist", A_CANT, 0);
}

void cu_do2pole(float *in, float *out, long n, t_twopole *x)
{
	float ym1 = x->c_ym1, ym2 = x->c_ym2, tmp;
	float a = x->c_coef1, b = x->c_coef2,
	g = x->c_gain;
	while (n--) {
		*out++ = g * (tmp = *in++ + a * ym1 + b * ym2);
		ym2 = ym1;
		ym1 = tmp;
	}
	x->c_ym1 = ym1;
	x->c_ym2 = ym2;
}

t_int *sigwahwah_perform(t_int *w)
{
	float *in, *cfreq, *q, *out;
	float theta, cos_theta, sin_theta, tmp, r;
	int n;
	t_sigwahwah *x;
	
	x = (t_sigwahwah *)(w[1]);
	if (x->x_obj.z_disabled)
		goto out;
	in = (float *)(w[2]);
	cfreq = (float *)(w[3]);
	q = (float *)(w[4]);
	out = (float *)(w[5]);
	n = w[6];
	
	theta = x->x_fconv * *cfreq;
	if (theta > 2.)
		theta = 2.;
	else if (theta < 0)
		theta = 0.;
	cos_theta = 1. - .5 * theta * theta;
	sin_theta = theta;
	if (sin_theta > 1.)
		sin_theta = 1.;
	tmp = *q;
	if (tmp <= .00001)
		r = 0.;
	else if ((r = 1. - theta/tmp) < 0)
		r = 0.;
	x->x_2p.c_gain = (1.-r) * ((1. - r) + r *sin_theta);
	x->x_2p.c_coef1 = 2.*r*cos_theta;
	x->x_2p.c_coef2 = -r*r;
	cu_do2pole(in, out, n, &x->x_2p);
out:
	return w + 7;
}

void sigwahwah_dsp(t_sigwahwah *x, t_signal **sp, short *count)
{
	x->x_2p.c_ym1 = 0.;
	x->x_2p.c_ym2 = 0.;
	x->x_2p.c_gain = 0.;
	x->x_fconv = (2. * 3.14159265) / sp[0]->s_sr;
	dsp_add(sigwahwah_perform, 6, x, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec,
		sp[3]->s_vec, sp[0]->s_n);
}

void sigwahwah_assist(t_sigwahwah *x, void *b, long m, long a, char *s)
{
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"(signal) input to be filtered"); break;
			case 1: sprintf(s,"(signal) Center Freq (Hz.)"); break;
			case 2: sprintf(s,"(signal) Q (resonance)"); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"(signal) filtered output"); break;
		}
	}
}

void *sigwahwah_new(long old)
{
	t_sigwahwah *x = (t_sigwahwah *)newobject(sigwahwah_class);
	dsp_setup((t_pxobject *)x,3);
	outlet_new((t_object *)x,"signal");
	return (x);
}

