// zerocross~ - based on Francois Dechelle's zerocross object
// MSP dudas hack 1998

// added report and clear messages for MSP rd 98

#include "ext.h"
#include "z_dsp.h"

void *zerocross_class;

typedef struct _zerocross
{
    t_pxobject x_obj;
	long x_count;
	double x_lastsample;
	void *x_out1;
} t_zerocross;

void *zerocross_new(long dummy);
t_int *zerocross_perform(t_int *w);
void zerocross_bang(t_zerocross *x);
void zerocross_report(t_zerocross *x);
void zerocross_clear(t_zerocross *x);
void zerocross_int(t_zerocross *x, long n);
void zerocross_assist(t_zerocross *x, void *b, long m, long a, char *s);
void zerocross_dsp(t_zerocross *x, t_signal **sp, short *count);

void main(void)
{
    setup((t_messlist **)&zerocross_class, (method)zerocross_new, (method)dsp_free, (short)sizeof(t_zerocross), 0L, 0L, 0);
    addmess((method)zerocross_dsp, "dsp", A_CANT, 0);	// dsp method
    addbang((method)zerocross_bang);
    addmess((method)zerocross_report, "report", 0);
    addmess((method)zerocross_clear, "clear", 0);
    addint((method)zerocross_int);
    
	addmess((method)zerocross_assist, "assist", A_CANT, 0);	

    dsp_initclass();								// must init the dsp
  
}

void zerocross_assist(t_zerocross *x, void *b, long m, long a, char *s)
{
	if (m==1) {
		switch (a) {
			case 0: sprintf(s,"Signal in, bang to report and clear zerocross count."); break;
		}
	} 
	else if (m==2) {
		switch (a) {
			case 0: sprintf(s,"Number of zero crosses since last bang."); break;
		}
	}
}

void *zerocross_new(long dummy)
{
    t_zerocross *x = (t_zerocross *)newobject(zerocross_class);
    dsp_setup((t_pxobject *)x,1);
	x->x_out1 = intout(x);
	x->x_count = 0;
	x->x_lastsample = 0.0;
    return (x);
}


void zerocross_bang(t_zerocross *x)
{
	outlet_int(x->x_out1, x->x_count);
	x->x_count = 0;
}

void zerocross_report(t_zerocross *x)
{
	outlet_int(x->x_out1, x->x_count);
}

void zerocross_clear(t_zerocross *x)
{
	x->x_count = 0;
}

void zerocross_int(t_zerocross *x, long n)
{
	if (n == 666)
		post("FTS is dead.",0);
}

t_int *zerocross_perform(t_int *w)		// if one sig inlet is used
{
	double lastsample, current;
	unsigned int lastsamplesign, currentsign;
	long count;
	
	t_zerocross *x = (t_zerocross *)(w[1]);
    t_float *in = (t_float *)(w[2]);
	int n = (int)(w[3]);			// vector size
	
	if (x->x_obj.z_disabled)
		goto out;

	lastsample = x->x_lastsample;
	lastsamplesign = (lastsample <= 0.0);
	count = x->x_count;
	
	while (--n) {
		current = *++in;
		currentsign = (current <= 0.0);
		count += lastsamplesign ^ currentsign;
		lastsamplesign = currentsign;
	}
	x->x_lastsample = current;
	x->x_count = count;
  
out:
    return (w+4);					// return pointer to NEXT object in dsp chain
}



// the dsp method that calls the above routines:

void zerocross_dsp(t_zerocross *x, t_signal **sp, short *count)  
{		
		dsp_add(zerocross_perform, 3, x, sp[0]->s_vec-1, sp[0]->s_n+1);
}

